"""
This module contains functions that can be used to clean EEG/MEG data using MNE-Python.
https://github.com/weiglszonja/meeg-tools/blob/master/README.md
"""
from random import sample
import numpy as np

from mne import Epochs
from mne.io import RawArray
from mne.preprocessing import bads, ICA
from mne.utils import logger
from autoreject import autoreject
from pyprep import NoisyChannels

from .utils.config import settings


def prepare_epochs_for_ica(epochs: Epochs) -> Epochs:
    """
    Drops epochs that were marked bad based on a global outlier detection.
    This implementation for the preliminary epoch rejection was based on the
    Python implementation of the FASTER algorithm from Marijn van Vliet
    https://gist.github.com/wmvanvliet/d883c3fe1402c7ced6fc
    Parameters
    ----------
    epochs

    Returns
    -------
    Epochs instance
    """
    logger.info("Preliminary epoch rejection: ")

    def _deviation(data: np.ndarray) -> np.ndarray:
        """
        Computes the deviation from mean for each channel.
        """
        channels_mean = np.mean(data, axis=2)
        return channels_mean - np.mean(channels_mean, axis=0)

    metrics = {
        "amplitude": lambda x: np.mean(np.ptp(x, axis=2), axis=1),
        "deviation": lambda x: np.mean(_deviation(x), axis=1),
        "variance": lambda x: np.mean(np.var(x, axis=2), axis=1),
    }

    epochs_data = epochs.get_data()

    bad_epochs = []
    for metric in metrics:
        scores = metrics[metric](epochs_data)
        outliers = bads._find_outliers(scores, threshold=3.0)
        logger.info(f"Bad epochs by {metric}\n\t{outliers}")
        bad_epochs.extend(outliers)

    bad_epochs = list(set(bad_epochs))
    epochs_faster = epochs.copy().drop(bad_epochs, reason="FASTER")

    return epochs_faster


def run_ica(epochs: Epochs, fit_params: dict = None) -> ICA:
    """
    Runs ICA decomposition on Epochs instance.

    If there are no EOG channels found, it tries to use 'Fp1' and 'Fp2' as EOG
    channels; if they are not found either, it chooses the first two channels
    to identify EOG components with mne.preprocessing.ica.find_bads_eog().
    Parameters
    ----------
    epochs: the instance to be used for ICA decomposition
    fit_params: parameters to be passed to ICA fit (e.g. orthogonal picard, extended infomax)
    Returns
    -------
    ICA instance
    """
    ica = ICA(
        n_components=settings["ica"]["n_components"],
        random_state=42,
        method=settings["ica"]["method"],
        fit_params=fit_params,
    )
    ica_epochs = epochs.copy()
    ica.fit(ica_epochs, decim=settings["ica"]["decim"])

    if "eog" not in epochs.get_channel_types():
        if "Fp1" and "Fp2" in epochs.get_montage().ch_names:
            eog_channels = ["Fp1", "Fp2"]
        else:
            eog_channels = epochs.get_montage().ch_names[:2]
        logger.info(
            "EOG channels are not found. Attempting to use "
            f'{",".join(eog_channels)} channels as EOG channels.'
        )
        ica_epochs.set_channel_types({ch: "eog" for ch in eog_channels})

    eog_indices, _ = ica.find_bads_eog(ica_epochs)
    ica.exclude = eog_indices

    return ica


def apply_ica(epochs: Epochs, ica: ICA) -> Epochs:
    """
    Applies ICA on Epochs instance.
    Parameters
    ----------
    epochs
    ica

    Returns
    -------

    """
    ica_epochs = epochs.copy().load_data()
    ica.apply(ica_epochs)

    ica_epochs.info.update(description=f"n_components: {len(ica.exclude)}")
    ica_epochs.info.update(temp=f'{epochs.info["temp"]}_ICA')

    return ica_epochs


def run_autoreject(
    epochs: Epochs, n_jobs: int = 11, subset: bool = False
) -> autoreject.RejectLog:
    """
    Finds bad epochs based on AutoReject.
    Parameters
    ----------
    epochs: the instance to be cleaned
    n_jobs: the number of parallel processes to be run
    subset: whether to train autoreject on a random subset of data (faster)

    Returns
    -------
    Autoreject instance
    """
    if not epochs.preload:
        epochs_autoreject = epochs.copy().load_data()
    else:
        epochs_autoreject = epochs.copy()

    ar = autoreject.AutoReject(random_state=42, n_jobs=n_jobs)

    n_epochs = len(epochs_autoreject)
    if subset:
        logger.info(
            f"Fitting autoreject on random (n={int(n_epochs * 0.25)}) "
            f"subset of epochs: "
        )
        subset = sample(set(np.arange(0, n_epochs, 1)), int(n_epochs * 0.25))
        ar.fit(epochs_autoreject[subset])

    else:
        logger.info(f"Fitting autoreject on (n={n_epochs}) epochs: ")
        ar.fit(epochs_autoreject)

    reject_log = ar.get_reject_log(epochs_autoreject)

    # report bad epochs where more than 15% (value updated from config.py)
    # of channels were marked as noisy within an epoch
    threshold = settings["autoreject"]["threshold"]
    num_bad_epochs = np.count_nonzero(reject_log.labels, axis=1)
    bad_epochs = np.where(num_bad_epochs > threshold * epochs.info["nchan"])
    bad_epochs = bad_epochs[0].tolist()

    reject_log.report = bad_epochs

    auto_bad_epochs = np.where(reject_log.bad_epochs)[0].tolist()
    if len(bad_epochs) < len(auto_bad_epochs):
        reject_log.report = sorted(list(set(bad_epochs + auto_bad_epochs)))

    logger.info(
        "\nAUTOREJECT report\n"
        f"There are {len(epochs_autoreject[reject_log.bad_epochs])} "
        f"bad epochs found with Autoreject. "
        f"You can assess these epochs with reject_log.bad_epochs\n"
        f"\nThere are {len(reject_log.report)} bad epochs where more than "
        f"{int(threshold * 100)}% of the channels were noisy. "
        f"You can assess these epochs with reject_log.report"
    )

    return reject_log


def apply_autoreject(epochs: Epochs, reject_log: autoreject.RejectLog) -> Epochs:
    """
    Drop bad epochs based on AutoReject.
    Parameters
    ----------
    epochs
    reject_log

    Returns
    -------

    """
    epochs_autoreject = epochs.copy().drop(reject_log.report, reason="AUTOREJECT")
    epochs_autoreject.info.update(temp=f'{epochs.info["temp"]}_autoreject')

    return epochs_autoreject


def get_noisy_channels(epochs: Epochs, with_ransac: bool = False) -> list:
    """
    Find bad channels using a range of methods as described in the PREP pipeline.
    Note that low-frequency trends should be removed from the EEG signal prior
    to bad channel detection.
    Read the documentation for further information about the methods:
    https://pyprep.readthedocs.io/en/latest/generated/pyprep.NoisyChannels.html#pyprep.NoisyChannels

    References
    ----------
    Bigdely-Shamlo, N., Mullen, T., Kothe, C., Su, K. M., Robbins, K. A. (2015).
    The PREP pipeline: standardized preprocessing for large-scale EEG analysis.
    Frontiers in Neuroinformatics, 9, 16.

    Parameters
    ----------
    epochs: Epochs object to use for bad channels detection
    with_ransac: whether RANSAC should be used for bad channel detection,
    in addition to the other methods.

    Returns
    -------
    list of bad channels names detected
    """
    # transform epochs to continuous data
    # to shape of (n_channels, n_epochs, n_times)
    data = np.transpose(epochs.get_data(), (1, 0, 2))
    # reshape to (n_channels, n_epochs * n_times) continuous data
    data = data.reshape((data.shape[0], data.shape[1] * data.shape[2]))
    # create Raw object from continuous data
    raw = RawArray(data=data, info=epochs.info)

    noisy_channels = NoisyChannels(raw=raw, do_detrend=False, random_state=42)
    noisy_channels.find_all_bads(ransac=with_ransac)

    bads = noisy_channels.get_bads(verbose=False)
    if bads:
        logger.info(
            "\nNoisyChannels REPORT\n"
            "------------------------"
            f"\n{np.round(len(bads) / len(epochs.ch_names), 2) * 100}%"
            f" of the channels were detected as noisy."
            f'\n({len(bads)}) channels: {", ".join(bads)}'
        )
    return bads


def interpolate_bad_channels(epochs: Epochs, bads: list) -> Epochs:
    """
    Interpolates channels in an Epochs instance based on a list of channel names.
    Parameters
    ----------
    epochs
    bads

    Returns
    -------

    """
    if not epochs.preload:
        epochs_interpolated = epochs.copy().load_data()
    else:
        epochs_interpolated = epochs.copy()

    epochs_interpolated.info["bads"] = bads

    if bads:
        bads_str = ", ".join(bads)
        description = f", interpolated: {bads_str}"
        epochs_interpolated.info.update(
            description=epochs.info["description"] + description
        )

    epochs_interpolated.interpolate_bads(reset_bads=True)
    epochs_interpolated.info.update(temp=f'{epochs.info["temp"]}_ransac')
    return epochs_interpolated



import os
from datetime import datetime
import re

import numpy as np
import pandas as pd

from mne import Epochs
from mne.utils import logger

from meeg_tools.utils.config import settings


def update_log(log_file_path: str, epochs: Epochs, notes: str) -> pd.DataFrame:
    """
    Documents the changes during preprocessing for an Epochs object.
    Custom description can be added with the notes argument.

    Parameters
    ----------
    log_file_path
    epochs
    notes

    Returns
    ----------
    log data
    """
    fid = epochs.info["temp"]
    dropped_epochs_marker = ["FASTER", "USER", "AUTOREJECT"]
    n_bad_epochs = len(
        [drop for drop in epochs.drop_log if np.isin(dropped_epochs_marker, drop).any()]
    )
    stimuli = list(epochs.event_id.keys())
    n_epochs_per_stimuli = [
        (", ").join([f"{ind}: {len(epochs[ind])}" for ind in stimuli])
    ]

    log = pd.DataFrame(
        {
            "fid": [fid],
            "highpass": [epochs.info["highpass"]],
            "lowpass": [epochs.info["lowpass"]],
            "n_components": [np.NaN],
            "n_bad_epochs": [n_bad_epochs],
            "total_drop_percentage": [round(epochs.drop_log_stats(), 2)],
            "n_epochs_per_stimuli": n_epochs_per_stimuli,
            "stimuli": [list(epochs.event_id.keys())],
            "t_min": [epochs.tmin],
            "t_max": [epochs.tmax],
            "n_interpolated": [np.NaN],
            "average_ref_applied": [bool(epochs.info["custom_ref_applied"])],
            "baseline": [epochs.baseline if epochs.baseline else np.NaN],
            "notes": [notes],
            "date_of_update": [datetime.utcnow().isoformat()],
            "author": [settings["log"]["author"]],
        }
    )

    description = epochs.info["description"]

    if description is None:
        logger.warning(
            'Failed to update parameters from epochs.info["description"], \n'
            "Returning log object, consider adding description field manually and rerunning this function.\n"
            'Formatting should match this format: "n_components: 1, interpolated: AF7, Fp2, P3, CP5, Fp1"'
        )
        return log

    if "n_components" in description:
        n_components = [x for x in description.split(",")[0] if x.isdigit()][0]
        log["n_components"].update(int(n_components))

    if "interpolated" in description:
        description_plain = description.replace(" ", "").replace(":", "")
        interpolated_channels_str = description_plain.split("interpolated")[1]
        n_interpolated = len(interpolated_channels_str.split(","))
        log["n_interpolated"].update(n_interpolated)
        log["interpolated"] = interpolated_channels_str

    author_clean = re.sub("\W+", "", settings["log"]["author"])
    log_file_name = f"{author_clean}_log.csv"
    if os.path.isfile(os.path.join(log_file_path, log_file_name)):
        log.to_csv(
            os.path.join(log_file_path, log_file_name),
            mode="a",
            index=False,
            header=False,
        )
    else:
        log.to_csv(os.path.join(log_file_path, log_file_name), index=False)

    return log



import os
import unittest

from mne import datasets


class TestRaw(unittest.TestCase):
    def setUp(self) -> None:
        self.raw_file_path = os.path.join(
            datasets.sample.data_path(), "MEG", "sample", "sample_audvis_raw.fif"
        )

    def test_read_raw_measurement(self) -> None:
        from meeg_tools.utils.raw import read_raw_measurement

        raw = read_raw_measurement(raw_file_path=self.raw_file_path)
        assert raw.info["fid"] == "sample_audvis_raw"
        assert bool(raw.get_montage())

    def test_filter_raw(self) -> None:
        from mne.io import read_raw
        from meeg_tools.utils.raw import filter_raw
        from meeg_tools.utils.config import settings

        settings["bandpass_filter"]["low_freq"] = 1
        settings["bandpass_filter"]["high_freq"] = 30

        raw = read_raw(fname=self.raw_file_path)

        raw_filtered = filter_raw(raw)

        assert raw_filtered.info["highpass"] == 1.0
        assert raw_filtered.info["lowpass"] == 30.0

        settings["bandpass_filter"]["low_freq"] = 5
        settings["bandpass_filter"]["high_freq"] = 45

        raw_filtered = filter_raw(raw)

        assert raw_filtered.info["highpass"] == 5.0
        assert raw_filtered.info["lowpass"] == 45.0

        assert raw.info["highpass"] != raw_filtered.info["highpass"]
        assert raw.info["lowpass"] != raw_filtered.info["lowpass"]



import os
from pathlib import Path

from mne import concatenate_raws
from mne.channels import read_custom_montage
from mne.io import read_raw, read_raw_edf, Raw
from mne.utils import logger

from .config import settings


def read_raw_measurement(raw_file_path: str, **kwargs):
    """
    Read raw EEG file from the given path.
    Parameters
    ----------
    raw_file_path: the full path to the EEG file
    locs_file_path: the full path to the channel locations file (optional)

    Returns
    -------
    Raw instance
    """
    raw_file_path = Path(raw_file_path)
    extension = raw_file_path.suffixes

    try:
        raw = read_raw(str(raw_file_path), preload=False, verbose=True)
    except Exception as e:
        if ".edf" in str(e):
            raw = read_raw_edf(str(raw_file_path), preload=False, verbose=True)
        else:
            logger.error(f"Unsupported file type ({extension})")

            return

    # Session parameters
    raw_id = raw_file_path.stem
    raw.info.update(temp=raw_id)

    if not bool(raw.get_montage()):
        logger.info(f"Channel locations are missing from the file")
        try:
            path_to_locs = kwargs["locs_file_path"]
            raw = set_raw_montage_from_locs(
                raw=raw, path_to_locs=path_to_locs, show_montage=False
            )
        except Exception as e:
            logger.error(e)
            return raw

    logger.info(raw.info)

    return raw


def set_raw_montage_from_locs(
    raw: Raw, path_to_locs: str, show_montage: False = bool
) -> Raw:
    """
    Reads channels locations from a file and applies them to Raw instance.
    Parameters
    ----------
    raw: the Raw instance with missing channel locations
    path_to_locs: the full path to the channel locations file
    (e.g. Starstim20.locs)
    show_montage: whether to show channel locations in a plot

    Returns
    -------
    Raw instance
    """
    if Path(path_to_locs).exists():
        montage = read_custom_montage(path_to_locs)

        # check if there are any channels not in raw instance
        missing_channel_locations = [
            ch_name for ch_name in raw.ch_names if ch_name not in montage.ch_names
        ]
        if missing_channel_locations:
            logger.info(
                f"There are {len(missing_channel_locations)} channel "
                f"positions not present in the "
                f"{Path(path_to_locs).stem} file."
            )
            logger.info(
                f"Assuming these ({missing_channel_locations}) "
                f"are not EEG channels, dropping them from Raw."
            )

            raw.drop_channels(missing_channel_locations)

        logger.info("Applying channel locations to Raw instance.")
        raw.set_montage(montage)

        if show_montage:
            raw.plot_sensors(show_names=True)

    else:
        logger.warning(
            f"Montage file {path_to_locs} path does not exist! "
            f"Returning unmodified raw."
        )

    return raw


def filter_raw(raw: Raw, **kwargs) -> Raw:
    """
    Applies band-pass filtering to Raw instance.
    Parameters
    ----------
    raw: the raw data to be filtered
    kwargs: optional keyword arguments to be passed to mne.filter()

    Returns
    -------
    Raw instance
    """
    raw_bandpass = (
        raw.load_data()
        .copy()
        .filter(
            l_freq=settings["bandpass_filter"]["low_freq"],
            h_freq=settings["bandpass_filter"]["high_freq"],
            n_jobs=8,
            **kwargs,
        )
    )

    return raw_bandpass


def concat_raws_with_suffix(path_to_raw_files: str, suffix: str) -> Raw:
    """
    Concatenates raw measurement files with a given suffix (e.g. ".vhdr") from a folder.
    File namings should follow an order e.g. the first part of the measurement is
    "eeg_1.vhdr" and the second part is "eeg_1_2.vhdr".
    Returns the concatenated instance as if it was a continuous measurement.
    Parameters
    ----------
    path_to_raw_files: str
        The path to the folder where the raw files are located.
    suffix: str
        The name of the file extension (e.g. ".vhdr", ".edf")
    Returns
    -------
    Raw instance
    """

    raw_file_path = Path(path_to_raw_files)
    file_names_in_order = sorted(
        [f.stem for f in raw_file_path.rglob("*.*") if f.suffix == suffix]
    )
    files = [
        str(f)
        for file_name in file_names_in_order
        for f in raw_file_path.rglob(f"{file_name}{suffix}")
    ]
    logger.info(
        f"Found {len(file_names_in_order)} files with {suffix} extension:\n"
        f"{', '.join(files)}"
    )
    raws = [read_raw(file, preload=False, verbose=True) for file in files]
    raw = concatenate_raws(raws)
    # Session parameters
    raw_id = file_names_in_order[0]
    raw.info.update(temp=raw_id)

    logger.info(raw.info)

    return raw



"""
This module contains functions to perform time-frequency analysis on EEG/MEG dataset.
"""
import os
from pathlib import Path
import re

import yaml
import pandas as pd
from tqdm import tqdm
from mne import read_epochs
from mne.utils import logger

from meeg_tools.connectivity import compute_connectivity
from meeg_tools.time_frequency import compute_power, save_to_hdf5, compute_erp

from meeg_tools.utils.epochs import create_metadata

# file path to configuration file
CONFIGURATION_FILE_PATH = "utils/config.yml"
# file path to preprocessed epochs
EPOCHS_FILE_PATH = "/Users/weian/Downloads/Raw_data/preprocessed/epochs_asrt_longer"
# postfix to lookup when collecting the files
EPOCHS_FILE_POSTFIX = "autoreject_ransac-epo.fif.gz"


def run():
    logger.info(f"\nReading configuration from {CONFIGURATION_FILE_PATH} file ...\n")
    config = read_config(CONFIGURATION_FILE_PATH)
    logger.info(yaml.dump(config))
    logger.info(
        f'\nRunning EEG/MEG {config["analysis"]["mode"]} analysis pipeline for '
        f'{config["analysis"]["picks"]} channels\n'
    )

    pipeline_in = Path(EPOCHS_FILE_PATH)
    current_folder_name = pipeline_in.name.replace("epochs",
                                                   config["analysis"]["mode"])
    if len(config["analysis"]["conditions"]):
        conditions = "-".join(config["analysis"]["conditions"]).upper()
        pipeline_out = (
                pipeline_in.parent
                / config["analysis"]["mode"]
                / current_folder_name
                / conditions
        )
    else:
        pipeline_out = (
                pipeline_in.parent
                / config["analysis"]["mode"]
                / current_folder_name
        )
    logger.info(f'Source directory is "{pipeline_in}"')
    logger.info(f'Output files are created at "{pipeline_out}"')

    if len(config["analysis"]["conditions"]):
        logger.info(
            f'Current run computes {config["analysis"]["mode"]} '
            f'for each {", ".join(config["analysis"]["conditions"])} '
            f"found in epochs metadata."
        )

    if not os.path.exists(pipeline_out):
        os.makedirs(pipeline_out)

    run_tfr_pipeline(source=pipeline_in, target=pipeline_out, conf=config)


def run_tfr_pipeline(source: Path, target: Path, conf: dict):
    def _file_exists(file_path: str) -> bool:
        return os.path.exists(file_path) and not conf["analysis"]["overwrite"]

    def _run_tfr(epochs_tfr):
        if mode == "power":
            power = compute_power(epochs=epochs_tfr, config=conf["power"])
            power.comment = file_path
            save_to_hdf5(power=power)
        elif mode == "erp":
            erp = compute_erp(epochs=epochs_tfr,
                              config=conf["erp"])
            erp.comment = str(file_path)
            erp.save(f"{file_path}_{mode}-{conf[mode]['postfix']}")
        elif mode == "con":
            con = compute_connectivity(epochs=epochs_tfr,
                                       config=conf["con"])
            con.attrs.update(comment=str(file_path),
                             ch_names=epochs_tfr.info["ch_names"],
                             sfreq=epochs_tfr.info["sfreq"],
                             ch_types=conf["analysis"]["picks"],)
            con.save(f"{file_path}_{conf[mode]['method']}-{conf[mode]['postfix']}")
        else:
            pass

    files = sorted(list(source.rglob(f"*{EPOCHS_FILE_POSTFIX}")))

    if not len(files):
        logger.warning(
            f"There are no files with the expected {EPOCHS_FILE_POSTFIX}."
            f"Doing nothing ..."
        )
        return

    condition_names = conf["analysis"]["conditions"]
    mode = conf["analysis"]["mode"]
    metadata = pd.DataFrame(columns=['fid', 'n_epochs'] + condition_names)
    pbar = tqdm(sorted(files))
    for file in pbar:
        pbar.set_description("Processing %s" % file.stem)
        fid = "_".join(str(file.stem.replace(" ", "_")).split("_")[:3])

        epochs = read_epochs(os.path.join(source, file), preload=False, verbose=0)
        epochs.load_data().pick_types(**{conf["analysis"]["picks"]: True})
        # make sure average ref is applied
        if not epochs.info["custom_ref_applied"]:
            epochs.load_data().set_eeg_reference("average")

        if condition_names:
            if isinstance(epochs.metadata, pd.DataFrame):
                epochs.metadata = create_metadata(epochs)
                epochs.metadata = epochs.metadata.astype(str)
                if all(name in epochs.metadata.columns for name in condition_names):
                    for condition_values, _ in epochs.metadata.groupby(condition_names):
                        if isinstance(condition_values, str):
                            condition_values = [condition_values]
                        query = [
                            f'{a} == "{b}"'
                            for a, b in zip(condition_names, condition_values)
                        ]
                        query_str = " & ".join(query)
                        epochs_query = epochs[query_str]

                        data = dict(fid=fid,
                                    n_epochs=len(epochs_query),
                                    **dict(zip(condition_names, condition_values)),
                                    **conf[mode])
                        metadata = metadata.append(data,
                                                   ignore_index=True)

                        file_name = f"{fid}_{'_'.join(condition_values)}"

                        floats = re.findall("[-+]?\d*\.\d+", file_name)
                        if floats:
                            for num in floats:
                                file_name = file_name.replace(num, str(int(float(num))))

                        file_path = target / file_name
                        if _file_exists(file_path=f"{file_path}_{mode}-{conf[mode]['postfix']}"):
                            pbar.update(1)
                            continue

                        _run_tfr(epochs_tfr=epochs_query)

        else:
            for event_id in epochs.event_id:
                epochs_per_event_id = epochs[event_id]

                metadata = metadata.append(dict(fid=fid,
                                                n_epochs=len(epochs_per_event_id),
                                                event_id=event_id,
                                                **conf[mode]),
                                           ignore_index=True)

                file_path = target / f"{fid}_{event_id}"
                if _file_exists(f"{file_path}_{mode}-{conf[mode]['postfix']}"):
                    pbar.update(1)
                    continue

                _run_tfr(epochs_tfr=epochs_per_event_id)

    pbar.close()

    metadata_file_path = os.path.join(target, f"{mode}_metadata.csv")
    metadata.to_csv(metadata_file_path, index=False)

    logger.info(f"Metadata file can be found at:\n{metadata_file_path}")

    logger.info("\n[PIPELINE FINISHED]")


def read_config(file_name: str):
    cfg = {}
    if os.path.exists(file_name):
        with open(file_name, "r") as config_file:
            cfg = yaml.safe_load(config_file)
    else:
        logger.error(f"Configuration file ({file_name}) does not exist!")
    return cfg


if __name__ == "__main__":
    run()



from typing import List
import numpy as np
import pandas as pd

from mne import (
    Epochs,
    find_events,
    events_from_annotations,
    make_fixed_length_events,
    merge_events,
    concatenate_epochs,
)
from mne.epochs import combine_event_ids
from mne.io import Raw
from mne.utils import logger

from .config import settings


def create_epochs_from_events(raw: Raw, event_ids: List) -> Epochs:
    """
    Create non-overlapping segments from Raw data.
    Note that temporal filtering should be done before creating the epochs.
    If there are annotations (triggers) found in the raw data, it creates
    epochs with respect to the stimulus onset defined by start_time
    and end_time in the configuration file (config.py) in seconds.
    Parameters
    ----------
    raw: the continuous data to be segmented into non-overlapping epochs
    event_ids: the list of event ids to create epochs from


    Returns
    -------
    Epochs instance
    """

    try:
        events = find_events(raw)
    except ValueError:
        events, _ = events_from_annotations(raw)

    selected_events = events[np.isin(events[..., 2], event_ids)]
    logger.info("Creating epochs from selected events ...")
    epochs = Epochs(
        raw=raw,
        events=selected_events,
        picks="all",
        event_id=list(np.unique(selected_events[..., 2])),
        baseline=None,
        tmin=settings["epochs"]["start_time"],
        tmax=settings["epochs"]["end_time"],
        preload=False,
    )

    return epochs


def create_epochs(raw: Raw) -> Epochs:
    """
    Create non-overlapping segments from Raw data with a fixed duration.
    Note that temporal filtering should be done before creating the epochs.
    The duration of epochs is defined in the configuration file (config.py).
    Parameters
    ----------
    raw: the continuous data to be segmented into non-overlapping epochs

    Returns
    -------
    Epochs instance
    """

    epoch_duration_in_seconds = settings["epochs"]["duration"]
    logger.info("Creating epochs from continuous data ...")
    events = make_fixed_length_events(
        raw, id=1, first_samp=True, duration=epoch_duration_in_seconds
    )

    epochs = Epochs(
        raw=raw,
        events=events,
        picks="all",
        event_id=list(np.unique(events[..., 2])),
        baseline=None,
        tmin=0.0,
        tmax=epoch_duration_in_seconds,  # - (1 / raw.info['sfreq']
        preload=False,
    )

    return epochs


def create_epochs_from_intervals(raw: Raw, intervals: List[tuple]) -> Epochs:
    events, _ = events_from_annotations(raw)

    epochs_list = []
    for interval in intervals:
        start_idx = np.where(events[..., 2] == interval[0])[0]
        end_idx = np.where(events[..., 2] == interval[1])[0]

        raw_cropped = raw.copy().crop(
            tmin=events[start_idx[0]][0] / raw.info["sfreq"],
            tmax=events[end_idx[0]][0] / raw.info["sfreq"],
        )

        epochs = create_epochs(raw_cropped)
        combine_event_ids(epochs, list(epochs.event_id.keys()), interval[0], copy=False)

        epochs_list.append(epochs)

    return concatenate_epochs(epochs_list)


def create_metadata(epochs: Epochs):
    """
    Custom function that adds metadata to Epochs instance.
    Parameters
    ----------
    epochs

    Returns
    -------

    """

    if isinstance(epochs.metadata, pd.DataFrame):
        metadata = epochs.metadata
    else:
        metadata = pd.DataFrame(
            data=epochs.events, columns=["time_in_samples", "stim", "id"]
        )

        # we can add boundaries of epochs
        epoch_boundaries = [211, 212, 213, 214, 215, 216]

        edges = np.where(np.isin(epochs.events[..., 2], epoch_boundaries))[0]

        boundaries = dict(zip(epoch_boundaries, edges))
        logger.info("Found these indices for these epoch boundary events: ")
        logger.info("\n".join("{}\t{}".format(k, v) for k, v in boundaries.items()))
        for epoch_ind, epoch in enumerate(np.split(epochs.events, edges, axis=0)):
            if epoch_ind != len(edges):
                metadata.loc[
                    metadata["time_in_samples"].isin(epoch[..., 0]), "epoch"
                ] = (epoch_ind + 1)

        metadata.loc[
            metadata["id"].isin([44, 45, 46, 47, 144, 145, 146, 147]), "answer"
        ] = "incorrect"
        metadata.loc[
            ~metadata["id"].isin([44, 45, 46, 47, 144, 145, 146, 147]), "answer"
        ] = "correct"

        # find stimuli that are followed by an incorrect answer
        stimuli = metadata[
            metadata["id"].isin(
                [10, 110, 11, 111, 12, 112, 13, 113, 14, 114, 15, 115, 16, 116]
            )
        ]
        stimuli_indices = np.asarray(stimuli["id"].index.tolist())
        incorrect_indices = (
            np.asarray(metadata.index[metadata["answer"] == "incorrect"].tolist()) - 1
        )
        incorrect_answers = stimuli_indices[np.isin(stimuli_indices, incorrect_indices)]
        metadata.loc[incorrect_answers, "answer"] = "incorrect"

    metadata.loc[
        metadata["id"].isin([10, 110, 11, 111, 14, 114, 15, 115]), "triplet"
    ] = "H"
    metadata.loc[
        metadata["id"].isin([12, 112, 13, 113, 16, 116, 112, 113, 116]), "triplet"
    ] = "L"
    metadata.loc[
        metadata["id"].isin(
            [10, 110, 14, 114],
        ),
        "triplet_type",
    ] = "HR"
    metadata.loc[
        metadata["id"].isin(
            [12, 112, 16, 116],
        ),
        "triplet_type",
    ] = "LR"
    metadata.loc[
        metadata["id"].isin(
            [11, 13, 15, 111, 113, 115],
        ),
        "triplet_type",
    ] = "P"

    metadata.loc[metadata["id"].isin([10, 11, 12, 13, 14, 15, 16]), "sequence"] = "A"
    metadata.loc[
        metadata["id"].isin([100, 111, 112, 113, 114, 115, 116]), "sequence"
    ] = "B"

    metadata.loc[metadata["id"].isin([11, 13, 15, 111, 113, 115]), "stimuli"] = "P"

    metadata.loc[
        metadata["id"].isin([10, 110, 14, 114, 12, 112, 16, 116]), "stimuli"
    ] = "R"

    metadata.loc[metadata["id"].isin([10, 11, 110, 111]), "rewiring"] = "H-L"

    metadata.loc[metadata["id"].isin([12, 13, 112, 113]), "rewiring"] = "L-H"

    metadata.loc[metadata["id"].isin([14, 15, 114, 115]), "rewiring"] = "H-H"

    metadata.loc[metadata["id"].isin([16, 116]), "rewiring"] = "L-L"

    return metadata


def create_new_events_from_metadata(epochs: Epochs) -> Epochs:
    # container for new events
    new_events_all = [np.zeros(3)]

    new_event_ids = {}
    event_counter = 0
    for triplet in epochs.metadata["triplet"].unique().tolist():
        for epoch in epochs.metadata["epoch"].unique().tolist():
            event_counter += 1
            epochs_filt = epochs[f"epoch == {int(epoch)} & triplet == '{triplet}'"]
            old_ids = list(epochs_filt.event_id.values())
            new_events = merge_events(
                events=epochs_filt.events,
                ids=old_ids,
                new_id=event_counter,
                replace_events=True,
            )
            new_event_ids[f"e{int(epoch)}_{triplet}"] = int(event_counter)
            new_events_all = np.concatenate([new_events_all, new_events], axis=0)

    new_events_all = new_events_all[1:]

    epochs.event_id = new_event_ids
    epochs.events = new_events_all
    return epochs



import os
import unittest

import numpy as np

from mne import datasets
from mne.io import read_raw


class TestEpochs(unittest.TestCase):
    def setUp(self) -> None:
        self.raw_file_path = os.path.join(
            datasets.sample.data_path(), "MEG", "sample", "sample_audvis_raw.fif"
        )

        self.raw = read_raw(fname=self.raw_file_path)

    def test_create_epochs_without_annotations(self):
        from meeg_tools.utils.epochs import create_epochs
        from meeg_tools.utils.config import settings

        settings["epochs"]["duration"] = 2

        epochs = create_epochs(self.raw)
        assert (epochs.events[1][0] - epochs.events[0][0]) == int(
            epochs.info["sfreq"] * settings["epochs"]["duration"]
        )

        epochs.load_data()
        # check epochs are not overlapping
        n_epochs = epochs.get_data().shape[0]
        for epoch_num in range(n_epochs - 1):
            assert not np.array_equal(
                epochs.get_data()[epoch_num, ...], epochs.get_data()[epoch_num + 1, ...]
            )

        assert int(len(self.raw) / self.raw.info["sfreq"] / 2) == len(epochs)
        assert epochs.event_id == {"1": 1}

    def test_create_epochs_from_events(self):
        from meeg_tools.utils.epochs import create_epochs_from_events
        from meeg_tools.utils.config import settings

        settings["epochs"]["start_time"] = -0.2
        settings["epochs"]["end_time"] = 0.5

        epochs = create_epochs_from_events(self.raw, event_ids=[1, 2, 3])
        epochs.load_data()

        # check epochs are not overlapping
        n_epochs = epochs.get_data().shape[0]
        for epoch_num in range(n_epochs - 1):
            assert not np.array_equal(
                epochs.get_data()[epoch_num, ...], epochs.get_data()[epoch_num + 1, ...]
            )

        assert len(epochs) == 218
        assert epochs.event_id == {"1": 1, "2": 2, "3": 3}
        assert np.round(epochs[0].times[0], 2) == -0.2
        assert np.round(epochs[0].times[-1], 2) == 0.5



import os
import unittest

from mne import datasets
from mne.io import read_raw

from meeg_tools.utils.epochs import create_epochs


class TestPreprocessing(unittest.TestCase):
    def setUp(self) -> None:
        self.raw_file_path = os.path.join(
            datasets.sample.data_path(), "MEG", "sample", "sample_audvis_raw.fif"
        )

        self.raw = read_raw(fname=self.raw_file_path)
        self.epochs = create_epochs(self.raw).load_data().pick_types(eeg=True)

    def test_prepare_epochs_for_ica(self):
        from meeg_tools.preprocessing import prepare_epochs_for_ica

        assert not any(self.epochs.drop_log)
        epochs_faster = prepare_epochs_for_ica(epochs=self.epochs)

        bad_epochs_indices = [
            drop_ind for drop_ind, drop in enumerate(epochs_faster.drop_log) if drop
        ]
        assert len(bad_epochs_indices) == 24
        assert bad_epochs_indices == [
            14,
            30,
            60,
            100,
            113,
            123,
            188,
            218,
            219,
            220,
            221,
            231,
            235,
            236,
            237,
            238,
            242,
            243,
            244,
            267,
            268,
            269,
            270,
            271,
        ]

    def test_run_ica(self):
        from meeg_tools.preprocessing import run_ica
        from meeg_tools.utils.config import settings

        settings["ica"]["n_components"] = 32
        settings["ica"]["decim"] = 3
        ica = run_ica(epochs=self.epochs)

        assert ica.exclude == [0]
        assert ica.n_components == 32
        assert "eog" not in self.epochs.get_channel_types()

    def test_run_autoreject(self):
        from meeg_tools.preprocessing import run_autoreject

        reject_log = run_autoreject(epochs=self.epochs, subset=False)
        epochs_autoreject = self.epochs.copy().drop(
            reject_log.report, reason="AUTOREJECT"
        )

        assert len(epochs_autoreject) < len(self.epochs)

    def test_get_noisy_channels(self):
        from meeg_tools.preprocessing import get_noisy_channels

        bads = get_noisy_channels(epochs=self.epochs, with_ransac=False)

        assert len(bads) == 9

    def test_interpolate_bad_channels(self):
        from meeg_tools.preprocessing import interpolate_bad_channels

        bads = ["EEG 025", "EEG 019"]

        epochs = interpolate_bad_channels(epochs=self.epochs, bads=bads)

        assert "EEG 025, EEG 019" in epochs.info["description"]

        epochs = interpolate_bad_channels(epochs=self.epochs, bads=[])

        assert "EEG 025, EEG 019" not in epochs.info["description"]



"""
This module contains functions that can be used to perform connectivity analysis
in the sensor space using MNE-Python.
https://mne.tools/mne-connectivity/stable/generated/mne_connectivity.spectral_connectivity.html
https://mne.tools/mne-connectivity/stable/auto_examples/sensor_connectivity.html#sphx-glr-auto-examples-sensor-connectivity-py
"""
from collections import Iterable

from matplotlib import pyplot as plt

from mne import Epochs
from mne_connectivity import spectral_connectivity_epochs, SpectralConnectivity
from mne.preprocessing import compute_current_source_density


def compute_connectivity(epochs: Epochs, config: dict) -> SpectralConnectivity:
    """
    Compute channel level connectivity matrix from Epochs instance.
    Returns the computed connectivity matrix (n_freqs, n_signals, n_signals).

    :return: np.ndarray con: The computed connectivity matrix with a shape of
    (n_freqs, n_signals, n_signals).

    See Also
    --------
    For frequency-decomposition and frequency bin reference:
    https://docs.scipy.org/doc/numpy/reference/generated/numpy.fft.rfftfreq.html
    """
    # spacing between frequency bins
    # spacing = np.round(epochs.info["sfreq"] / epochs.get_data().shape[-1], 2)
    freq_bands = sorted(config["bands"].values())

    low_cutoff = tuple(freq[0] for freq in freq_bands)
    high_cutoff = tuple(freq[1] for freq in freq_bands)
    # high_cutoff = tuple(
    #     freq[1] - spacing if freq != max(freq_bands) else freq[1]
    #     for freq in freq_bands
    # )

    if config["use_laplace"]:
        epochs_csd = compute_surface_laplacian(epochs=epochs, verbose=False)
    else:
        epochs_csd = epochs.copy()

    con = spectral_connectivity_epochs(
        epochs_csd,
        method=config["method"],
        mode='multitaper',
        sfreq=epochs.info["sfreq"],
        fmin=low_cutoff,
        fmax=high_cutoff,
        faverage=True,
        tmin=config["tmin"],
        tmax=config["tmax"],
        n_jobs=11,
        verbose=False)
    # con, _, _, _, _ = spectral_connectivity(
    #     data=epochs_csd,
    #     method=kwargs["method"],
    #     sfreq=epochs.info["sfreq"],
    #     mode=kwargs["mode"],
    #     fmin=low_cutoff,
    #     fmax=high_cutoff,
    #     faverage=kwargs["faverage"],
    #     n_jobs=kwargs["n_jobs"],
    #     verbose=True,
    # )
    #
    # # from shape of (n_signals, n_signals, n_freqs) to
    # # (n_freqs, n_signals, n_signals)
    # con = np.transpose(con, (2, 0, 1))
    # con = abs(con)

    if all(isinstance(el, Iterable) for el in con.attrs["freqs_used"]):
        con.attrs["freqs_used"] = [freq for freq_arr in con.attrs["freqs_used"]
                                   for freq in freq_arr]

    # make sure all connectivity values are positive
    con.xarray.values = abs(con.xarray.values)

    return con


def compute_surface_laplacian(epochs: Epochs, verbose: bool = True) -> Epochs:
    """
    Performs the surface Laplacian transform on the Epochs instance
    For more information about the transform parameters please refer to
    Cohen, M. X. (2014). Analyzing neural time series data: theory and practice
    . MIT press. For more information about this function in particular,
    visit the MNE documentation at:
    https://mne.tools/dev/generated/mne.preprocessing.compute_current_source_density.html
    Parameters
    ----------
    epochs: the epochs to be transformed
    verbose: whether to visualize the power spectral densities before and after
    the Laplacian transform

    Returns
    -------
    Raw instance
    """

    epochs_csd = compute_current_source_density(epochs.copy())

    if verbose:
        fig, axes_subplot = plt.subplots(
            nrows=2, ncols=1, sharex="all", sharey="all", dpi=220
        )

        epochs.plot_psd(ax=axes_subplot[0], show=False, fmax=60)
        epochs_csd.plot_psd(ax=axes_subplot[1], show=False, fmax=60)
        fig.show()

    return epochs_csd



"""
This module contains functions that can be used to perform time-frequency analysis on
EEG/MEG data with MNE-Python.
"""
import os
import re
from typing import List
from pathlib import Path

import numpy as np
import pandas as pd

from mne import Epochs, EvokedArray, pick_channels, read_evokeds, Evoked
from mne.channels import combine_channels, find_ch_adjacency
from mne.stats import combine_adjacency, permutation_cluster_1samp_test
from mne.time_frequency import tfr_morlet, AverageTFR, read_tfrs
from mne.utils import logger
from yasa import irasa


def compute_power(epochs: Epochs, config: dict) -> AverageTFR:
    """
    Computes Time-Frequency Representation (TFR) using complex Morlet wavelets
    averaged over epochs. Power is written to an HDF5 file.
    Parameters
    ----------
    epochs
    config
    """
    fmin = config["morlet"]["fmin"]
    fmax = config["morlet"]["fmax"]
    step = config["morlet"]["step"]
    freqs = np.logspace(*np.log10([fmin, fmax]), num=step)
    n_cycles = freqs / 2.0

    power = tfr_morlet(
        epochs.average() if config["is_evoked"] else epochs.copy(),
        freqs=freqs,
        n_cycles=n_cycles,
        use_fft=True,
        return_itc=False,
        decim=config["morlet"]["decim"],
        average=True,
        n_jobs=-1,
    )

    return power


def save_to_hdf5(power: AverageTFR):
    if not os.path.exists(power.comment.parent):
        os.makedirs(power.comment.parent)

    power.comment = str(power.comment)
    # replace floats in file name with integers
    floats = re.findall("[-+]?\d*\.\d+", power.comment)
    if floats:
        for num in floats:
            power.comment = power.comment.replace(num, str(int(float(num))))
    file_path_with_extension = f"{power.comment}_power-tfr.h5"

    logger.info(f"Saving power at {file_path_with_extension} ...")
    power.save(fname=str(file_path_with_extension), overwrite=True)
    logger.info("[FINISHED]")


def average_power_into_frequency_bands(power: AverageTFR, config: dict) -> AverageTFR:
    """
    Computes average power in specific frequency ranges defined in config
    Parameters
    ----------
    power
    config

    Returns
    -------

    """
    band_values = config.values()
    band_power_data_arr = []

    for band in band_values:
        band_filter = np.logical_and(
            power.freqs >= float(band[0]), power.freqs < float(band[1])
        )
        if int(max(power.freqs)) == int(max(band_values)[1]):
            band_filter[-1] = True
        band_data = power.data[:, band_filter, :].mean(axis=1)
        band_power_data_arr.append(band_data[np.newaxis, :])

    band_power_data = np.concatenate(band_power_data_arr, axis=0)
    band_power_data = np.transpose(band_power_data, (1, 0, 2))

    band_freqs = np.asarray([band[0] for band in band_values])

    band_power = power.copy()
    band_power.data = band_power_data
    band_power.freqs = band_freqs
    band_power.comment = power.comment + "_band_average"

    return band_power


def compute_power_irasa_method(epochs: Epochs, bands: tuple):
    """
    Computes power by separating the aperiodic (= fractal, or 1/f) and oscillatory
    component of the power spectra of EEG data using the IRASA method.
    https://raphaelvallat.com/yasa/build/html/generated/yasa.irasa.html
    Parameters
    ----------
    epochs
    bands

    Returns
    -------

    """
    # transform epochs to continuous data
    # to shape of (n_channels, n_epochs, n_times)
    data = np.transpose(epochs.get_data(), (1, 0, 2))
    # reshape to (n_channels, n_epochs * n_times) continuous data
    data = data.reshape((data.shape[0], data.shape[1] * data.shape[2]))

    freqs, _, psd_oscillatory, _ = irasa(
        data=data,
        sf=epochs.info["sfreq"],
        ch_names=epochs.info["ch_names"],
        band=bands,
        return_fit=True,
        win_sec=2 * (1 / bands[0]),
    )

    return freqs, psd_oscillatory


def compute_erp(epochs: Epochs, config: dict) -> EvokedArray:
    evoked = epochs.average(method=config["method"])

    return evoked


def get_erp_peak_measures(erp: EvokedArray,
                          tmin: float,
                          tmax: float,
                          mode: str,
                          picks=None,
                          combine='mean') -> pd.DataFrame:
    """
    Computes peak measures (peak latency, peak amplitude) from Evoked instance for a
    given time interval defined by tmin and tmax in seconds. Peak measures can be
    computed for ROIs (averages data over given list of channels) by defining the list
    of channel names (e.g. ['F2', 'F5', 'F7']) that will be aggregated first,
    and then the measures computed.
    If picks is not defined, peak measures are computed for each channel.
    Parameters
    ----------
    erp
    tmin
    tmax
    mode: 'pos': finds the peak with a positive voltage (ignores negative voltages)
    'neg': finds the peak with a negative voltage (ignores positive voltages)
    'abs': finds the peak with the largest absolute voltage regardless of sign (positive or negative)
    combine: whether to combine channels (defined by picks) using 'mean', 'median' or 'std'
    use None to not combine channels but get peak measures separately for each channel
    picks

    Returns
    -------
    DataFrame containing the peak measures for each channel or for a given ROI
    """
    if picks is None:
        picks = []

    erp_measures = pd.DataFrame(
        columns=[
            "fid",
            "ch_name",
            "tmin",
            "tmax",
            "mode",
            "peak_latency",
            "peak_amplitude",
        ]
    )

    if combine:
        picks_idx = pick_channels(erp.info["ch_names"], include=picks)
        roi_erp = combine_channels(erp, dict(roi=picks_idx), method=combine)
        _, lat, amp = roi_erp.get_peak(ch_type='eeg',
                                       tmin=tmin,
                                       tmax=tmax,
                                       mode=mode,
                                       return_amplitude=True)

        ch_name = ' '.join(picks)
        erp_measures = erp_measures.append(dict(fid=erp.comment,
                                                ch_name=ch_name,
                                                tmin=tmin,
                                                tmax=tmax,
                                                mode=mode,
                                                peak_latency=lat,
                                                peak_amplitude=amp * 1e6),
                                           ignore_index=True)

    else:
        for ch_name in picks:
            _, lat, amp = erp.copy().pick(ch_name).get_peak(ch_type='eeg',
                                                            tmin=tmin,
                                                            tmax=tmax,
                                                            mode=mode,
                                                            return_amplitude=True)

            erp_measures = erp_measures.append(dict(fid=erp.comment,
                                                    ch_name=ch_name,
                                                    tmin=tmin,
                                                    tmax=tmax,
                                                    mode=mode,
                                                    peak_latency=lat,
                                                    peak_amplitude=amp * 1e6),
                                               ignore_index=True)

    return erp_measures


def get_mean_amplitude(erp, tmin, tmax, mode):
    data = erp.crop(tmin=tmin, tmax=tmax).data
    sign_mean_data = data.squeeze()
    if mode == "pos":
        if not np.any(data > 0):
            logger.warning(
                f"{erp.comment} :" "No positive values encountered. Using default mode."
            )
        else:
            sign_mean_data = data[data > 0]
    elif mode == "neg":
        if not np.any(data < 0):
            logger.warning(
                f"{erp.comment} :" "No negative values encountered. Using default mode."
            )
        else:
            sign_mean_data = data[data < 0]
    elif mode == "abs":
        sign_mean_data = abs(sign_mean_data)

    mean_amplitude = sign_mean_data.mean(axis=0) * 1e6

    return mean_amplitude


def get_erp_measures_from_cross_condition_data(
    erp_arrays: List[EvokedArray],
    cross_condition_data: pd.DataFrame,
    interval_in_seconds: float,
):
    erp_measures = pd.DataFrame(
        columns=[
            "fid",
            "ch_name",
            "tmin",
            "tmax",
            "mode",
            "peak_latency",
            "peak_amplitude",
            "mean_amplitude",
        ]
    )

    picks = cross_condition_data['ch_name'].values[0].split()
    mode = cross_condition_data['mode'].values[0]
    for erp in erp_arrays:
        picks_idx = pick_channels(erp.info['ch_names'], include=picks)

        if len(picks_idx) > 1:
            roi_erp = combine_channels(erp, dict(roi=picks_idx), method='mean')
            tmin = cross_condition_data['peak_latency'].values[0] - interval_in_seconds
            tmax = cross_condition_data['peak_latency'].values[0] + interval_in_seconds

            if tmax > erp.tmax:
                tmax = erp.tmax
            if tmin < erp.tmin:
                tmin = erp.tmin

            _, lat, amp = roi_erp.get_peak(ch_type='eeg',
                                           tmin=tmin,
                                           tmax=tmax,
                                           mode=mode,
                                           return_amplitude=True)

            mean_amp = get_mean_amplitude(erp=roi_erp, tmin=tmin, tmax=tmax, mode=mode)

            fid = Path(erp.comment.replace("\\", "/")).name
            ch_name = cross_condition_data['ch_name'].values[0]
            erp_measures = erp_measures.append(dict(fid=fid,
                                                    ch_name=ch_name,
                                                    tmin=tmin,
                                                    tmax=tmax,
                                                    mode=mode,
                                                    peak_latency=lat,
                                                    peak_amplitude=amp * 1e6,
                                                    mean_amplitude=mean_amp),
                                               ignore_index=True)

        else:
            for ch_name in cross_condition_data['ch_name']:
                tmin = cross_condition_data[
                           cross_condition_data['ch_name'] == ch_name][
                           'peak_latency'] - interval_in_seconds
                tmax = cross_condition_data[
                           cross_condition_data['ch_name'] == ch_name][
                           'peak_latency'] + interval_in_seconds

                tmin = tmin.values[0]
                tmax = tmax.values[0]

                if tmax > erp.tmax:
                    tmax = erp.tmax
                if tmin < erp.tmin:
                    tmin = erp.tmin

                _, lat, amp = erp.copy().pick(ch_name).get_peak(ch_type='eeg',
                                                                tmin=tmin,
                                                                tmax=tmax,
                                                                mode=mode,
                                                                return_amplitude=True)

                mean_amp = get_mean_amplitude(erp=erp.copy().pick(ch_name),
                                              tmin=tmin,
                                              tmax=tmax,
                                              mode=mode)

                fid = Path(erp.comment.replace("\\", "/")).name
                erp_measures = erp_measures.append(dict(fid=fid,
                                                        ch_name=ch_name,
                                                        tmin=tmin,
                                                        tmax=tmax,
                                                        mode=mode,
                                                        peak_latency=lat,
                                                        peak_amplitude=amp * 1e6,
                                                        mean_amplitude=mean_amp),
                                                   ignore_index=True)

    return erp_measures


def read_tfrs_from_path(tfrs_path: str, pattern: str) -> List[AverageTFR]:
    """
    Reads TFR files from path with a given pattern to look for within the file name. I.e.
    used for separating conditions (e.g. H, L).
    Parameters
    ----------
    tfrs_path
    pattern

    Returns
    -------

    """
    files = sorted(
        [
            f
            for f in os.listdir(tfrs_path)
            if pattern.lower() in f.lower() and not f.startswith(".")
        ]
    )
    tfrs = [read_tfrs(os.path.join(tfrs_path, f))[0] for f in files]

    return tfrs


def read_evokeds_from_path(evokeds_path: str, pattern: str) -> List[Evoked]:
    """
    Reads Evoked files from path with a given pattern to look for within the file name. I.e.
    used for separating conditions (e.g. H, L).
    Parameters
    ----------
    evokeds_path
    pattern

    Returns
    -------

    """
    files = sorted(
        [
            f
            for f in os.listdir(evokeds_path)
            if pattern.lower() in f.lower() and not f.startswith(".")
        ]
    )
    evoked = [read_evokeds(os.path.join(evokeds_path, f), verbose=0)[0] for f in files]

    return evoked


def compute_power_difference(
    power_condition1: List[AverageTFR],
    power_condition2: List[AverageTFR],
    picks: [],
    baseline: None,
    tmin: float,
    tmax: float,
):

    if not picks:
        picks = None

    diff_over_participants = []
    for power1, power2 in zip(power_condition1, power_condition2):
        power1.pick_channels(picks).apply_baseline(baseline).crop(tmin=tmin, tmax=tmax)
        power2.pick_channels(picks).apply_baseline(baseline).crop(tmin=tmin, tmax=tmax)
        difference_data = (power1.data * 1e6) - (power2.data * 1e6)

        diff_over_participants.append(difference_data[np.newaxis, ...])

    diff_data = np.concatenate(diff_over_participants, axis=0)

    return diff_data


def permutation_correlation(diff_data, info, n_permutations, p_value):
    sensor_adjacency, ch_names = find_ch_adjacency(info, "eeg")
    adjacency = combine_adjacency(
        sensor_adjacency, diff_data.shape[2], diff_data.shape[3]
    )

    T_obs, clusters, cluster_p_values, H0 = permutation_cluster_1samp_test(
        diff_data,
        n_permutations=n_permutations,
        threshold=None,
        tail=0,
        adjacency=adjacency,
        out_type="mask",
        verbose=True,
    )

    # Create new stats image with only significant clusters for plotting
    T_obs_plot = np.nan * np.ones_like(T_obs)
    for c, p_val in zip(clusters, cluster_p_values):
        if p_val <= p_value:
            print(f"Significant cluster with p-value {p_val}")
            T_obs_plot[c] = T_obs[c]

    return T_obs_plot, cluster_p_values[cluster_p_values <= p_value]



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages


with open('README.md', 'r') as f:
    long_description = f.read()

# Get requirements
with open('requirements.txt') as f:
    install_requires = f.read().strip().split('\n')

setup_args = dict(
    name='meeg-tools',
    version='0.3.4',
    description='EEG/MEEG data preprocessing and analyses tools',
    python_requires='<3.10.1',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Szonja Weigl',
    author_email='weigl.anna.szonja@gmail.com',
    url='https://github.com/weiglszonja/meeg-tools',
    packages=find_packages(),
    include_package_data=True,
    install_requires=install_requires,
)


if __name__ == '__main__':
    setup(**setup_args)



settings = dict(
    log=dict(author="Szonja Weigl"),
    bandpass_filter=dict(low_freq=0.5, high_freq=45),
    epochs=dict(start_time=-0.700, end_time=0.750, duration=1),
    ica=dict(n_components=32, method="picard", decim=None),
    autoreject=dict(threshold=0.15),
)
